{llvm::Intrinsic::bitreverse, ::mlir::LLVM::BitReverseOp::getOperationName()},
{llvm::Intrinsic::copysign, ::mlir::LLVM::CopySignOp::getOperationName()},
{llvm::Intrinsic::coro_align, ::mlir::LLVM::CoroAlignOp::getOperationName()},
{llvm::Intrinsic::coro_begin, ::mlir::LLVM::CoroBeginOp::getOperationName()},
{llvm::Intrinsic::coro_end, ::mlir::LLVM::CoroEndOp::getOperationName()},
{llvm::Intrinsic::coro_free, ::mlir::LLVM::CoroFreeOp::getOperationName()},
{llvm::Intrinsic::coro_id, ::mlir::LLVM::CoroIdOp::getOperationName()},
{llvm::Intrinsic::coro_resume, ::mlir::LLVM::CoroResumeOp::getOperationName()},
{llvm::Intrinsic::coro_save, ::mlir::LLVM::CoroSaveOp::getOperationName()},
{llvm::Intrinsic::coro_size, ::mlir::LLVM::CoroSizeOp::getOperationName()},
{llvm::Intrinsic::coro_suspend, ::mlir::LLVM::CoroSuspendOp::getOperationName()},
{llvm::Intrinsic::cos, ::mlir::LLVM::CosOp::getOperationName()},
{llvm::Intrinsic::ctlz, ::mlir::LLVM::CountLeadingZerosOp::getOperationName()},
{llvm::Intrinsic::cttz, ::mlir::LLVM::CountTrailingZerosOp::getOperationName()},
{llvm::Intrinsic::ctpop, ::mlir::LLVM::CtPopOp::getOperationName()},
{llvm::Intrinsic::eh_typeid_for, ::mlir::LLVM::EhTypeidForOp::getOperationName()},
{llvm::Intrinsic::exp2, ::mlir::LLVM::Exp2Op::getOperationName()},
{llvm::Intrinsic::exp, ::mlir::LLVM::ExpOp::getOperationName()},
{llvm::Intrinsic::fabs, ::mlir::LLVM::FAbsOp::getOperationName()},
{llvm::Intrinsic::ceil, ::mlir::LLVM::FCeilOp::getOperationName()},
{llvm::Intrinsic::floor, ::mlir::LLVM::FFloorOp::getOperationName()},
{llvm::Intrinsic::fma, ::mlir::LLVM::FMAOp::getOperationName()},
{llvm::Intrinsic::fmuladd, ::mlir::LLVM::FMulAddOp::getOperationName()},
{llvm::Intrinsic::get_active_lane_mask, ::mlir::LLVM::GetActiveLaneMaskOp::getOperationName()},
{llvm::Intrinsic::log10, ::mlir::LLVM::Log10Op::getOperationName()},
{llvm::Intrinsic::log2, ::mlir::LLVM::Log2Op::getOperationName()},
{llvm::Intrinsic::log, ::mlir::LLVM::LogOp::getOperationName()},
{llvm::Intrinsic::maxnum, ::mlir::LLVM::MaxNumOp::getOperationName()},
{llvm::Intrinsic::maximum, ::mlir::LLVM::MaximumOp::getOperationName()},
{llvm::Intrinsic::memcpy_inline, ::mlir::LLVM::MemcpyInlineOp::getOperationName()},
{llvm::Intrinsic::memcpy, ::mlir::LLVM::MemcpyOp::getOperationName()},
{llvm::Intrinsic::memmove, ::mlir::LLVM::MemmoveOp::getOperationName()},
{llvm::Intrinsic::memset, ::mlir::LLVM::MemsetOp::getOperationName()},
{llvm::Intrinsic::minnum, ::mlir::LLVM::MinNumOp::getOperationName()},
{llvm::Intrinsic::minimum, ::mlir::LLVM::MinimumOp::getOperationName()},
{llvm::Intrinsic::powi, ::mlir::LLVM::PowIOp::getOperationName()},
{llvm::Intrinsic::pow, ::mlir::LLVM::PowOp::getOperationName()},
{llvm::Intrinsic::prefetch, ::mlir::LLVM::Prefetch::getOperationName()},
{llvm::Intrinsic::round, ::mlir::LLVM::RoundOp::getOperationName()},
{llvm::Intrinsic::sadd_with_overflow, ::mlir::LLVM::SAddWithOverflowOp::getOperationName()},
{llvm::Intrinsic::smax, ::mlir::LLVM::SMaxOp::getOperationName()},
{llvm::Intrinsic::smin, ::mlir::LLVM::SMinOp::getOperationName()},
{llvm::Intrinsic::smul_with_overflow, ::mlir::LLVM::SMulWithOverflowOp::getOperationName()},
{llvm::Intrinsic::ssub_with_overflow, ::mlir::LLVM::SSubWithOverflowOp::getOperationName()},
{llvm::Intrinsic::sin, ::mlir::LLVM::SinOp::getOperationName()},
{llvm::Intrinsic::sqrt, ::mlir::LLVM::SqrtOp::getOperationName()},
{llvm::Intrinsic::stackrestore, ::mlir::LLVM::StackRestoreOp::getOperationName()},
{llvm::Intrinsic::stacksave, ::mlir::LLVM::StackSaveOp::getOperationName()},
{llvm::Intrinsic::experimental_stepvector, ::mlir::LLVM::StepVectorOp::getOperationName()},
{llvm::Intrinsic::uadd_with_overflow, ::mlir::LLVM::UAddWithOverflowOp::getOperationName()},
{llvm::Intrinsic::umax, ::mlir::LLVM::UMaxOp::getOperationName()},
{llvm::Intrinsic::umin, ::mlir::LLVM::UMinOp::getOperationName()},
{llvm::Intrinsic::umul_with_overflow, ::mlir::LLVM::UMulWithOverflowOp::getOperationName()},
{llvm::Intrinsic::usub_with_overflow, ::mlir::LLVM::USubWithOverflowOp::getOperationName()},
{llvm::Intrinsic::vp_ashr, ::mlir::LLVM::VPAShrOp::getOperationName()},
{llvm::Intrinsic::vp_add, ::mlir::LLVM::VPAddOp::getOperationName()},
{llvm::Intrinsic::vp_and, ::mlir::LLVM::VPAndOp::getOperationName()},
{llvm::Intrinsic::vp_fadd, ::mlir::LLVM::VPFAddOp::getOperationName()},
{llvm::Intrinsic::vp_fdiv, ::mlir::LLVM::VPFDivOp::getOperationName()},
{llvm::Intrinsic::vp_fmul, ::mlir::LLVM::VPFMulOp::getOperationName()},
{llvm::Intrinsic::vp_fneg, ::mlir::LLVM::VPFNegOp::getOperationName()},
{llvm::Intrinsic::vp_fpext, ::mlir::LLVM::VPFPExtOp::getOperationName()},
{llvm::Intrinsic::vp_fptosi, ::mlir::LLVM::VPFPToSIOp::getOperationName()},
{llvm::Intrinsic::vp_fptoui, ::mlir::LLVM::VPFPToUIOp::getOperationName()},
{llvm::Intrinsic::vp_fptrunc, ::mlir::LLVM::VPFPTruncOp::getOperationName()},
{llvm::Intrinsic::vp_frem, ::mlir::LLVM::VPFRemOp::getOperationName()},
{llvm::Intrinsic::vp_fsub, ::mlir::LLVM::VPFSubOp::getOperationName()},
{llvm::Intrinsic::vp_fma, ::mlir::LLVM::VPFmaOp::getOperationName()},
{llvm::Intrinsic::vp_inttoptr, ::mlir::LLVM::VPIntToPtrOp::getOperationName()},
{llvm::Intrinsic::vp_lshr, ::mlir::LLVM::VPLShrOp::getOperationName()},
{llvm::Intrinsic::vp_load, ::mlir::LLVM::VPLoadOp::getOperationName()},
{llvm::Intrinsic::vp_merge, ::mlir::LLVM::VPMergeMinOp::getOperationName()},
{llvm::Intrinsic::vp_mul, ::mlir::LLVM::VPMulOp::getOperationName()},
{llvm::Intrinsic::vp_or, ::mlir::LLVM::VPOrOp::getOperationName()},
{llvm::Intrinsic::vp_ptrtoint, ::mlir::LLVM::VPPtrToIntOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_add, ::mlir::LLVM::VPReduceAddOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_and, ::mlir::LLVM::VPReduceAndOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_fadd, ::mlir::LLVM::VPReduceFAddOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_fmax, ::mlir::LLVM::VPReduceFMaxOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_fmin, ::mlir::LLVM::VPReduceFMinOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_fmul, ::mlir::LLVM::VPReduceFMulOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_mul, ::mlir::LLVM::VPReduceMulOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_or, ::mlir::LLVM::VPReduceOrOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_smax, ::mlir::LLVM::VPReduceSMaxOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_smin, ::mlir::LLVM::VPReduceSMinOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_umax, ::mlir::LLVM::VPReduceUMaxOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_umin, ::mlir::LLVM::VPReduceUMinOp::getOperationName()},
{llvm::Intrinsic::vp_reduce_xor, ::mlir::LLVM::VPReduceXorOp::getOperationName()},
{llvm::Intrinsic::vp_sdiv, ::mlir::LLVM::VPSDivOp::getOperationName()},
{llvm::Intrinsic::vp_sext, ::mlir::LLVM::VPSExtOp::getOperationName()},
{llvm::Intrinsic::vp_sitofp, ::mlir::LLVM::VPSIToFPOp::getOperationName()},
{llvm::Intrinsic::vp_srem, ::mlir::LLVM::VPSRemOp::getOperationName()},
{llvm::Intrinsic::vp_select, ::mlir::LLVM::VPSelectMinOp::getOperationName()},
{llvm::Intrinsic::vp_shl, ::mlir::LLVM::VPShlOp::getOperationName()},
{llvm::Intrinsic::vp_store, ::mlir::LLVM::VPStoreOp::getOperationName()},
{llvm::Intrinsic::experimental_vp_strided_load, ::mlir::LLVM::VPStridedLoadOp::getOperationName()},
{llvm::Intrinsic::experimental_vp_strided_store, ::mlir::LLVM::VPStridedStoreOp::getOperationName()},
{llvm::Intrinsic::vp_sub, ::mlir::LLVM::VPSubOp::getOperationName()},
{llvm::Intrinsic::vp_trunc, ::mlir::LLVM::VPTruncOp::getOperationName()},
{llvm::Intrinsic::vp_udiv, ::mlir::LLVM::VPUDivOp::getOperationName()},
{llvm::Intrinsic::vp_uitofp, ::mlir::LLVM::VPUIToFPOp::getOperationName()},
{llvm::Intrinsic::vp_urem, ::mlir::LLVM::VPURemOp::getOperationName()},
{llvm::Intrinsic::vp_xor, ::mlir::LLVM::VPXorOp::getOperationName()},
{llvm::Intrinsic::vp_zext, ::mlir::LLVM::VPZExtOp::getOperationName()},
{llvm::Intrinsic::vacopy, ::mlir::LLVM::VaCopyOp::getOperationName()},
{llvm::Intrinsic::vaend, ::mlir::LLVM::VaEndOp::getOperationName()},
{llvm::Intrinsic::vastart, ::mlir::LLVM::VaStartOp::getOperationName()},
{llvm::Intrinsic::masked_compressstore, ::mlir::LLVM::masked_compressstore::getOperationName()},
{llvm::Intrinsic::masked_expandload, ::mlir::LLVM::masked_expandload::getOperationName()},
{llvm::Intrinsic::vector_reduce_add, ::mlir::LLVM::vector_reduce_add::getOperationName()},
{llvm::Intrinsic::vector_reduce_and, ::mlir::LLVM::vector_reduce_and::getOperationName()},
{llvm::Intrinsic::vector_reduce_fmax, ::mlir::LLVM::vector_reduce_fmax::getOperationName()},
{llvm::Intrinsic::vector_reduce_fmin, ::mlir::LLVM::vector_reduce_fmin::getOperationName()},
{llvm::Intrinsic::vector_reduce_mul, ::mlir::LLVM::vector_reduce_mul::getOperationName()},
{llvm::Intrinsic::vector_reduce_or, ::mlir::LLVM::vector_reduce_or::getOperationName()},
{llvm::Intrinsic::vector_reduce_smax, ::mlir::LLVM::vector_reduce_smax::getOperationName()},
{llvm::Intrinsic::vector_reduce_smin, ::mlir::LLVM::vector_reduce_smin::getOperationName()},
{llvm::Intrinsic::vector_reduce_umax, ::mlir::LLVM::vector_reduce_umax::getOperationName()},
{llvm::Intrinsic::vector_reduce_umin, ::mlir::LLVM::vector_reduce_umin::getOperationName()},
{llvm::Intrinsic::vector_reduce_xor, ::mlir::LLVM::vector_reduce_xor::getOperationName()},
{llvm::Intrinsic::vscale, ::mlir::LLVM::vscale::getOperationName()},
